program hexdump;
{$APPTYPE CONSOLE}
uses
	Windows, SysUtils, Registry;

	{*******************}
	{ Buffered file i/o }
	{*******************}

Var
	{* Files *}
	fileInput : File;
	fileOutput : File;

Const
	rbsMax = 64000;
	wbsMax = 64000;

Type
	TBuffer = Array [ 0 .. rbsMax - 1 ] Of Byte;
	PBuffer = ^TBuffer;

Var
	rb, wb : PBuffer;
	rbs, rbc, wbs, wbc : Longint; { buffer size / buffer count }

	AllReaded, EOIF : Boolean;

{*******}
{ Opens }
{*******}

Function openInput ( FileName : String ) : Byte;
var
	io_Result : Integer;
Begin
	{ init }
	AllReaded := False;
	EOIF := False;
	rb := Nil;
	rbs := 0;
	rbc := 0;

	{ attempt to open }
	Assign ( fileInput, FileName );
	FileMode := fmOpenRead; 
	{$I-}
	Reset ( fileInput, 1 );
	{$I+}
	io_Result := IOResult;
	If io_Result <> 0
	Then Begin
		openInput := io_Result;
	End
	Else Begin
		openInput := 0;
	End;
End;

Procedure CloseInput;
Begin
	CloseFile ( fileInput );
End;

{***********************}
{* Grabs one byte from *}
{* the input stream    *}
{***********************}

Function GetByte : Byte;
Begin
	If EOIF
	Then Begin
		GetByte := 0;
	End
	Else Begin
		If rbc = rbs
		Then Begin
			If AllReaded
			Then Begin
				EOIF := True;
			End
			Else Begin
				If rbs > 0
				Then Begin
					FreeMem ( rb, rbs );
				End;
				rbs := FileSize ( fileInput ) - FilePos ( fileInput );
				If rbs > rbsMax
				Then Begin
					rbs := rbsMax;
				End;
				(*
				If rbs > MaxAvail
				Then Begin
					{ allocate half of the mem for reading }
					rbs := MaxAvail Shr 1 And $FF00;
				End;
        *)
				GetMem ( rb, rbs );
				BlockRead ( fileInput, rb^, rbs );
				rbc := 0;
				If FilePos ( fileInput ) = FileSize ( fileInput )
				Then Begin
					AllReaded := True;
				End;
			End;
		End;
		GetByte := rb^ [ rbc ];
		Inc ( rbc );

		If ( rbc = rbs ) And AllReaded
		Then Begin
			EOIF := True;
		End;
	End;
End;

Function GetLine : String;
Var
	temp : String;
	b : Byte;
Begin
	temp := '';
	b := GetByte;
	While ( Not EOIF ) And ( b <> 10 ) Do Begin
		If ( b <> 13 )
		Then temp := temp + Chr ( b );
		b := GetByte;
	End;
	GetLine := temp;
End;

{*******}
{ Opens }
{*******}

Function writeOutput ( FileName : String ) : Byte;
Var
	io_Result : Integer;
Begin
	{ init }
	wb := Nil;
	wbs := 0;
	wbc := 0;

	{ attempt to open }
	AssignFile ( fileOutput, FileName );
	{$I-}
	ReWrite ( fileOutput, 1 );
	{$I+}
	io_Result := IOResult;
	If io_Result <> 0
	Then Begin
		writeOutput := io_Result;
	End
	Else Begin
		writeOutput := 0;
	End;
End;

Procedure Flush;
Begin
	{ is there anything to flush? }
	If ( wbc > 0 )
	Then Begin
		BlockWrite ( fileOutput, wb^ , wbc );

		{ Throw stuff away }
		FreeMem ( wb, wbs );
		wbs := 0;
		wbc := 0;
	End;
End;

Procedure sendString ( s : String );
Var
	l : Byte;
Begin
	{ Do we need to allocate a new buffer? }
	If wbs = 0
	Then Begin
		{ Yeah. }
		(*
		wbs := MaxAvail Shr 1 And $FF00;
		If wbs > wbsMax
		Then Begin
		*)
			wbs := wbsMax;
			(*
		End;
		*)
		GetMem ( wb, wbs );
		wbc := 0;
	End;
	{ Push the stuff into the buffer }
	l := Length ( s );
	If l > 0
	Then Begin
		Move ( s [ 1 ], wb^ [ wbc ], l );
		Inc ( wbc, l );
	End;
	wb^ [ wbc ] := 13;
	Inc ( wbc );
	wb^ [ wbc ] := 10;
	Inc ( wbc );
	{ Is it gonna be full? }
	If wbc + $100 > wbs
	Then Begin
		{ Yeah, let's flush it! }
		Flush;
	End;
End;

Procedure sendByte ( b : Byte );
Begin
	{ Do we need to allocate a new buffer? }
	If wbs = 0
	Then Begin
		{ Yeah. }
		(*
		wbs := MaxAvail Shr 1 And $FF00;
		If wbs > wbsMax
		Then Begin
		*)
			wbs := wbsMax;
			(*
		End;
		*)
		GetMem ( wb, wbs );
		wbc := 0;
	End;
	{ Push the stuff into the buffer }
	wb^ [ wbc ] := b;
	Inc ( wbc );

	{ Is it gonna be full? }
	If wbc + $100 > wbs
	Then Begin
		{ Yeah, let's flush it! }
		Flush;
	End;
End;

Procedure CloseOutput;
Begin
	Flush;
	CloseFile ( fileOutput );
End;

	{*****}
	{ Hex }
	{*****}

Const
	HexDigits: Array [ 00..15 ] Of Char =
	( '0', '1', '2', '3', '4', '5', '6', '7',
		'8', '9', 'A', 'B', 'C', 'D', 'E', 'F' );

Function Hex ( i : Longint; Digits : Byte ) : String;
Var
	SHex : String [ 8 ];
	CharCounter : Byte;
Begin
	SHex := '';
	Digits := ( Digits - 1 ) And 7;
	For CharCounter := 0 To Digits Do Begin
		SHex := HexDigits [ i And $F ] + SHex;
		i := i Shr 4;
	End;
	Hex := SHex;
End;

Function HexB ( i : Byte ) : String;
Begin
	HexB := HexDigits [ i Shr 4 ] + HexDigits [ i And $F ];
End;

Function ChangeExt ( sPath, sExt : String ) : String;
Var
	i : Integer;
	b : Boolean;
Begin
	i := Length ( sPath );
	b := False;
	While ( i > 0 ) And Not b Do
	Begin
		If sPath [ i ] = '.' Then
			b := True
		Else
			If sPath [ i ] = '\' Then
			Begin
				b := True;
				i := Length ( sPath );
			End
			Else
				Dec ( i );
	End;
	If i = 0 Then i := Length ( sPath ) + 1;
	ChangeExt := Copy ( sPath, 1, i - 1 ) + sExt;
End;

procedure DoRegister();
Var
	r : TRegistry;
	sCommand : String;
Begin
	r := TRegistry.Create;
	r.RootKey := HKEY_CLASSES_ROOT;
	//
	If Not r.OpenKey ( '*\shell', False ) Then
	Begin
		r.CreateKey ( '*\shell' );
	End;
	r.CloseKey;
	//
	sCommand := 'hexdump';
	//
	If Not r.OpenKey ( '*\shell\' + sCommand, False ) Then
	Begin
		r.CreateKey ( '*\shell\' + sCommand );
		r.OpenKey ( '*\shell\' + sCommand, False );
		r.WriteString ( '', 'HexDump' );
		r.CloseKey;
		//
		r.CreateKey ( '*\shell\' + sCommand + '\command' );
	End
	Else
	Begin
		r.CloseKey;
	End;
	//
	r.OpenKey ( '*\shell\' + sCommand + '\command', True );
	r.WriteString ( '', ParamStr ( 0 ) + ' "%1"' );
	r.CloseKey;
	//
	r.Free;
end;

procedure DoUnRegister();
Var
	r : TRegistry;
	sCommand : String;
Begin
	r := TRegistry.Create;
	r.RootKey := HKEY_CLASSES_ROOT;
	//
	sCommand := 'hexdump';
	//
	If r.OpenKey ( '*\shell\' + sCommand, False ) Then
	Begin
		r.CloseKey;
		//
		r.DeleteKey ( '*\shell\' + sCommand );
	End
	Else
	Begin
		r.CloseKey;
	End;
	//
	r.Free;
end;

var
	b : Byte;
	s : String;
	c, c2 : Byte;

Begin
	If ParamCount = 0 Then
	Begin
		WriteLn ( 'Missing filename.' );
		WriteLn ( 'Usage: ' + ParamStr ( 0 ) + ' <filename>' );
		WriteLn ( '    or ' + ParamStr ( 0 ) + ' -reg (to register a custom HexDump command)' );
		WriteLn ( '       ' + ParamStr ( 0 ) + ' -unreg (to unregister it)' );
		ReadLn;
		Halt ( 1 );
	End
	Else
	Begin
		If UpperCase ( ParamStr ( 1 ) ) = '-REG' Then
		Begin
			DoRegister;
			WriteLn ( 'Custom command registered.' );
		End
		Else
		Begin
			If UpperCase ( ParamStr ( 1 ) ) = '-UNREG' Then
			Begin
				DoUnRegister;
				WriteLn ( 'Custom command unregistered.' );
			End
			Else
			Begin
				If OpenInput ( ParamStr ( 1 ) ) <> 0 Then
				Begin
					WriteLn ( 'Could not open "' + ParamStr ( 1 ) + '".' );
					Halt ( 1 );
				End
				Else
				Begin
					WriteOutput ( ChangeExt ( ParamStr ( 1 ), '_hex.asm' ) );

					c := 0;
					c2 := 0;

					If UpperCase ( Copy ( ParamStr ( 1 ), Length ( ParamStr ( 1 ) ) - 2, 3 ) ) = 'PRG' Then
					Begin
						GetByte; { skip PRG loading address }
						GetByte;
					End;

					While Not Eoif Do
					Begin
						b := GetByte;
						if c = 0 Then s := #9 + 'db ';
						s := s + '$' + HexB ( b );
						if ( c < 15 ) And Not Eoif Then s := s + ',';
						Inc ( c );
						If c = 16 Then
						Begin
							SendString ( s );
							s := '';
							c := 0;
							Inc ( c2 );
							If c2 = 16 Then
							Begin
								SendString ( s );
								c2 := 0;
							End;
						End;
					End;
					If s <> '' Then SendString ( s );

					CloseOutput;
					CloseInput;
				End;
			End;
		End;
	End;
End.
